#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import TYPE_CHECKING

import pytest
from argilla_server.apis.v0.models.commons.model import BulkResponse
from argilla_server.apis.v0.models.text_classification import (
    TextClassificationAnnotation,
    TextClassificationBulkRequest,
    TextClassificationQuery,
    TextClassificationRecord,
    TextClassificationSearchRequest,
    TextClassificationSearchResults,
)
from argilla_server.commons.models import PredictionStatus, TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User
from argilla_server.schemas.v0.datasets import Dataset

if TYPE_CHECKING:
    from httpx import AsyncClient


@pytest.mark.asyncio
async def test_create_records_for_text_classification_with_multi_label(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_create_records_for_text_classification_with_multi_label"

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    records = [
        TextClassificationRecord.parse_obj(data)
        for data in [
            {
                "id": 0,
                "inputs": {"data": "my data"},
                "multi_label": True,
                "metadata": {
                    "field_one": "value one",
                    "field_two": "value 2",
                    "one_more": [{"a": 1, "b": 2}],
                },
                "prediction": {
                    "agent": "testA",
                    "labels": [
                        {"class": "Test", "score": 0.6},
                        {"class": "Mocking", "score": 0.7},
                        {"class": "NoClass", "score": 0.2},
                    ],
                },
            },
            {
                "id": 1,
                "inputs": {"data": "my data"},
                "multi_label": True,
                "metadata": {
                    "field_one": "another value one",
                    "field_two": "value 2",
                },
                "prediction": {
                    "agent": "testB",
                    "labels": [
                        {"class": "Test", "score": 0.6},
                        {"class": "Mocking", "score": 0.7},
                        {"class": "NoClass", "score": 0.2},
                    ],
                },
            },
        ]
    ]
    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == 2

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            tags={"new": "tag"},
            metadata={"new": {"metadata": "value"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    response = await async_client.get(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200, response.json()

    dataset_response = Dataset.parse_obj(response.json())
    assert dataset_response.tags == {"env": "test", "class": "text classification", "new": "tag"}
    assert dataset_response.metadata == {"config": {"the": "config"}, "new": {"metadata": "value"}}

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search", json={}, params=workspace_query_params
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == 2
    assert results.aggregations.predicted_as == {"Mocking": 2, "Test": 2}
    assert results.records[0].predicted is None

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search",
        json={"query": {"predicted_by": ["testA"]}},
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == len(results.records) == 1
    assert results.aggregations.predicted_by == {"testA": 1}


@pytest.mark.asyncio
async def test_create_records_for_text_classification(async_client: "AsyncClient", argilla_user: User, test_telemetry):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_create_records_for_text_classification"

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    tags = {"env": "test", "class": "text classification"}
    metadata = {"config": {"the": "config"}}
    classification_bulk = TextClassificationBulkRequest(
        tags=tags,
        metadata=metadata,
        records=[
            TextClassificationRecord(
                **{
                    "id": 0,
                    "inputs": {"data": "my data"},
                    "prediction": {
                        "agent": "test",
                        "labels": [
                            {"class": "Test", "score": 0.3},
                            {"class": "Mocking", "score": 0.7},
                        ],
                    },
                }
            )
        ],
    )

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=classification_bulk.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == 1

    response = await async_client.get(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    created_dataset = Dataset.parse_obj(response.json())
    assert created_dataset.tags == tags
    assert created_dataset.metadata == metadata

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search", json={}, params=workspace_query_params
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == 1
    assert results.aggregations.dict(exclude={"score"}) == {
        "annotated_as": {},
        "annotated_by": {},
        "metadata": {},
        "predicted": {},
        "predicted_as": {"Mocking": 1},
        "predicted_by": {"test": 1},
        "status": {"Default": 1},
        "words": {"data": 1},
    }

    test_telemetry.track_data.assert_called()


@pytest.mark.asyncio
async def test_create_records_for_text_classification_vector_search(
    async_client: "AsyncClient", argilla_user: User, test_telemetry
):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_create_records_for_text_classification_vector_search"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    tags = {"env": "test", "class": "text classification"}
    metadata = {"config": {"the": "config"}}
    classification_bulk = TextClassificationBulkRequest(
        tags=tags,
        metadata=metadata,
        records=[
            TextClassificationRecord(**a)
            for a in [
                {
                    "id": 0,
                    "inputs": {"data": "my data"},
                    "prediction": {
                        "agent": "test",
                        "labels": [
                            {"class": "Test", "score": 0.3},
                            {"class": "Mocking", "score": 0.7},
                        ],
                    },
                    "vectors": {"my_bert": {"value": [10, 11, 12, 13]}},
                },
                {
                    "id": 1,
                    "inputs": {"data": "your data"},
                    "prediction": {
                        "agent": "test",
                        "labels": [
                            {"class": "Test", "score": 0.35},
                            {"class": "Mocking", "score": 0.65},
                        ],
                    },
                    "vectors": {"my_bert": {"value": [14, 15, 16, 17]}},
                },
                {
                    "id": 2,
                    "inputs": {"data": "their data"},
                    "prediction": {
                        "agent": "test",
                        "labels": [
                            {"class": "Test", "score": 0.4},
                            {"class": "Mocking", "score": 0.6},
                        ],
                    },
                    "vectors": {"my_bert": {"value": [14, 15, 16, 18]}},
                },
            ]
        ],
    )
    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=classification_bulk.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == 3

    response = await async_client.get(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    created_dataset = Dataset.parse_obj(response.json())
    assert created_dataset.tags == tags
    assert created_dataset.metadata == metadata

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search", json={}, params=workspace_query_params
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == 3
    assert all(hasattr(record, "vectors") for record in results.records)
    assert results.aggregations.dict(exclude={"score"}) == {
        "annotated_as": {},
        "annotated_by": {},
        "metadata": {},
        "predicted": {},
        "predicted_as": {"Mocking": 3},
        "predicted_by": {"test": 3},
        "status": {"Default": 3},
        "words": {"data": 3},
    }

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search",
        json={
            "query": {
                "vector": {
                    "name": "my_bert",
                    "value": [14, 15, 16, 17],
                },
            }
        },
        params=workspace_query_params,
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == 3
    assert all(hasattr(record, "vectors") for record in results.records)
    # similarity ordered records
    assert [record.id for record in results.records] == [1, 2, 0]


@pytest.mark.asyncio
async def test_partial_record_update(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    name = "test_partial_record_update"
    response = await async_client.delete(f"/api/datasets/{name}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": name, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    record = TextClassificationRecord(
        **{
            "id": 1,
            "inputs": {"text": "This is a text, oh yeah!"},
            "prediction": {
                "agent": "test",
                "labels": [
                    {"class": "Positive", "score": 0.6},
                    {"class": "Negative", "score": 0.3},
                    {"class": "Other", "score": 0.1},
                ],
            },
        }
    )

    bulk = TextClassificationBulkRequest(records=[record])

    response = await async_client.post(
        f"/api/datasets/{name}/TextClassification:bulk",
        json=bulk.dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.failed == 0
    assert bulk_response.processed == 1

    record.annotation = TextClassificationAnnotation.parse_obj(
        {"agent": "gold_standard", "labels": [{"class": "Positive"}]}
    )

    bulk.records = [record]
    await async_client.post(
        f"/api/datasets/{name}/TextClassification:bulk", json=bulk.dict(by_alias=True), params=workspace_query_params
    )

    response = await async_client.post(
        f"/api/datasets/{name}/TextClassification:search",
        json={"query": TextClassificationQuery(predicted=PredictionStatus.OK).dict(by_alias=True)},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == 1

    first_record = results.records[0]
    assert first_record.last_updated is not None

    first_record.last_updated = None
    assert TextClassificationRecord(**first_record.dict(by_alias=True, exclude_none=True)) == TextClassificationRecord(
        **{
            "id": 1,
            "inputs": {"text": "This is a text, oh yeah!"},
            "prediction": {
                "agent": "test",
                "labels": [
                    {"class": "Positive", "score": 0.6},
                    {"class": "Negative", "score": 0.3},
                    {"class": "Other", "score": 0.1},
                ],
            },
            "annotation": {
                "agent": "gold_standard",
                "labels": [{"class": "Positive"}],
            },
        }
    )


@pytest.mark.asyncio
async def test_sort_by_last_updated(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_sort_by_last_updated"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    for i in range(0, 10):
        response = await async_client.post(
            f"/api/datasets/{dataset}/TextClassification:bulk",
            json=TextClassificationBulkRequest(
                records=[
                    TextClassificationRecord(
                        **{
                            "id": i,
                            "inputs": {"data": "my data"},
                            "metadata": {"s": "value"},
                        }
                    )
                ],
            ).dict(by_alias=True),
            params=workspace_query_params,
        )
        assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search?from=0&limit=10",
        json={"sort": [{"id": "last_updated", "order": "asc"}]},
        params=workspace_query_params,
    )
    assert response.status_code == 200
    assert [r["id"] for r in response.json()["records"]] == list(range(0, 10))


@pytest.mark.asyncio
async def test_sort_by_id_as_default(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_sort_by_id_as_default"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            records=[
                TextClassificationRecord(
                    **{
                        "id": i,
                        "inputs": {"data": "my data"},
                        "metadata": {"s": "value"},
                    }
                )
                for i in range(0, 100)
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search?from=0&limit=10", json={}, params=workspace_query_params
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == 100
    assert list(map(lambda r: r.id, results.records)) == [0, 1, 10, 11, 12, 13, 14, 15, 16, 17]


@pytest.mark.asyncio
async def test_some_sort_by(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_some_sort_by"
    expected_records_length = 50

    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            records=[
                TextClassificationRecord(
                    **{
                        "id": i,
                        "inputs": {"data": "my data"},
                        "prediction": {"agent": f"agent_{i%5}", "labels": []},
                        "metadata": {
                            "s": f"{i} value",
                        },
                    }
                )
                for i in range(0, expected_records_length)
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search?from=0&limit=10",
        json={
            "sort": [
                {"id": "wrong_field"},
            ]
        },
        params=workspace_query_params,
    )
    assert response.status_code == 400

    expected_response_property_name_2_value = {
        "detail": {
            "code": "argilla.api.errors::BadRequestError",
            "params": {
                "message": "Wrong sort id wrong_field. Valid values "
                "are: ['id', 'metadata', 'score', "
                "'predicted', 'predicted_as', "
                "'predicted_by', 'annotated_as', "
                "'annotated_by', 'status', 'last_updated', "
                "'event_timestamp']"
            },
        }
    }
    assert response.json()["detail"]["code"] == expected_response_property_name_2_value["detail"]["code"]
    assert (
        response.json()["detail"]["params"]["message"]
        == expected_response_property_name_2_value["detail"]["params"]["message"]
    )
    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search?from=0&limit=10",
        json={
            "sort": [
                {"id": "predicted_by", "order": "desc"},
                {"id": "metadata.s", "order": "asc"},
            ]
        },
        params=workspace_query_params,
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == expected_records_length
    assert list(map(lambda r: r.id, results.records)) == [14, 19, 24, 29, 34, 39, 4, 44, 49, 9]


@pytest.mark.asyncio
async def test_disable_aggregations_when_scroll(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_disable_aggregations_when_scroll"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)

    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=[
                TextClassificationRecord(
                    **{
                        "id": i,
                        "inputs": {"data": "my data"},
                        "prediction": {
                            "agent": "test",
                            "labels": [
                                {"class": "Test", "score": 0.3},
                                {"class": "Mocking", "score": 0.7},
                            ],
                        },
                    }
                )
                for i in range(0, 100)
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.processed == 100

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search?from=10",
        json={},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == 100
    assert results.aggregations is None


@pytest.mark.asyncio
async def test_include_event_timestamp(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_include_event_timestamp"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200
    request = TextClassificationBulkRequest(
        tags={"env": "test", "class": "text classification"},
        metadata={"config": {"the": "config"}},
        records=[
            TextClassificationRecord(
                **{
                    "id": i,
                    "inputs": {"data": "my data"},
                    "event_timestamp": datetime.utcnow(),
                    "prediction": {
                        "agent": "test",
                        "labels": [{"class": "Test", "score": 0.3}, {"class": "Mocking", "score": 0.7}],
                    },
                }
            )
            for i in range(0, 100)
        ],
    )
    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        # Use content instead of json to properly serialize event_timestamp datetime
        content=request.json(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.processed == 100

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search?from=10", json={}, params=workspace_query_params
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.total == 100
    assert all(map(lambda record: record.event_timestamp is not None, results.records))


@pytest.mark.asyncio
async def test_words_cloud(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_language_detection"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)

    assert response.status_code == 200
    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            records=[
                TextClassificationRecord(**{"id": 0, "inputs": {"text": "Esto es un ejemplo de texto"}}),
                TextClassificationRecord(**{"id": 1, "inputs": {"text": "This is an simple text example"}}),
                TextClassificationRecord(**{"id": 2, "inputs": {"text": "C'est nes pas une pipe"}}),
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    BulkResponse.parse_obj(response.json())
    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search", json={}, params=workspace_query_params
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert results.aggregations.words is not None


@pytest.mark.asyncio
async def test_metadata_with_point_in_field_name(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_metadata_with_point_in_field_name"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            records=[
                TextClassificationRecord(
                    **{
                        "id": 0,
                        "inputs": {"text": "Esto es un ejemplo de texto"},
                        "metadata": {"field.one": 1, "field.two": 2},
                    }
                ),
                TextClassificationRecord(
                    **{
                        "id": 1,
                        "inputs": {"text": "This is an simple text example"},
                        "metadata": {"field.one": 1, "field.two": 2},
                    }
                ),
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search?limit=0", json={}, params=workspace_query_params
    )
    assert response.status_code == 200

    results = TextClassificationSearchResults.parse_obj(response.json())
    assert "field.one" in results.aggregations.metadata
    assert results.aggregations.metadata.get("field.one", {})["1"] == 2
    assert results.aggregations.metadata.get("field.two", {})["2"] == 2


@pytest.mark.asyncio
async def test_wrong_text_query(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_wrong_text_query"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            records=[
                TextClassificationRecord(
                    **{
                        "id": 0,
                        "inputs": {"text": "Esto es un ejemplo de texto"},
                        "metadata": {"field.one": 1, "field.two": 2},
                    }
                ),
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search",
        json=TextClassificationSearchRequest(query=TextClassificationQuery(query_text="!")).dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 400

    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::InvalidTextSearchError",
            "params": {"message": "Failed to parse query [!]"},
        }
    }


@pytest.mark.asyncio
async def test_search_using_text(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_search_using_text"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets",
        json={"name": dataset, "task": TaskType.text_classification, "workspace": argilla_user.username},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:bulk",
        json=TextClassificationBulkRequest(
            records=[
                TextClassificationRecord(
                    **{
                        "id": 0,
                        "inputs": {"data": "Esto es un ejemplo de Texto"},
                        "metadata": {"field.one": 1, "field.two": 2},
                    }
                ),
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search",
        json=TextClassificationSearchRequest(query=TextClassificationQuery(query_text="text: texto")).dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    assert response.json()["total"] == 1

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search",
        json=TextClassificationSearchRequest(query=TextClassificationQuery(query_text="text.exact: texto")).dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    assert response.json()["total"] == 0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import uuid4

from typer import Typer
from typer.testing import CliRunner


class TestCliServerSearchEngineReindex:
    # TODO: This test should create multiple datasets and records so they are reindexed.
    # In order to do that right now we need to duplicate all async test factories to be synchronous.
    # Instead of do that we can (once we move away from asynchronous requests) use regular factories here and improve
    # the test.
    def test_reindex(self, cli_runner: CliRunner, cli: Typer):
        result = cli_runner.invoke(cli, "search-engine reindex")

        assert result.exit_code == 0, result.output

    def test_reindex_with_nonexistent_dataset_id(self, cli_runner: CliRunner, cli: Typer):
        result = cli_runner.invoke(cli, f"search-engine reindex --feedback-dataset-id {uuid4()}")

        assert result.exit_code == 1



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from argilla_server.apis.v0.models.text2text import (
    Text2TextAnnotation,
    Text2TextPrediction,
    Text2TextQuery,
    Text2TextRecord,
)
from argilla_server.daos.backend.search.query_builder import EsQueryBuilder


def test_sentences_sorted_by_score():
    record = Text2TextRecord(
        text="The inpu2 text",
        prediction=Text2TextAnnotation(
            agent="test_sentences_sorted_by_score",
            sentences=[
                Text2TextPrediction(text="sentence 1", score=0.6),
                Text2TextPrediction(text="sentence 2", score=0.5),
                Text2TextPrediction(text="sentence 3", score=1.0),
            ],
        ),
    )

    assert record.prediction.sentences[0].score == 1.0
    assert record.prediction.sentences[1].score == 0.6
    assert record.prediction.sentences[2].score == 0.5


def test_model_dict():
    record = Text2TextRecord(
        id=0,
        text="The input text",
        prediction=Text2TextAnnotation(
            agent="test_sentences_sorted_by_score",
            sentences=[
                Text2TextPrediction(text="sentence 1", score=0.6),
                Text2TextPrediction(text="sentence 2", score=0.5),
                Text2TextPrediction(text="sentence 3", score=1.0),
            ],
        ),
    )
    assert record.dict(exclude_none=True) == {
        "id": 0,
        "metrics": {},
        "prediction": {
            "agent": "test_sentences_sorted_by_score",
            "sentences": [
                {"score": 1.0, "text": "sentence 3"},
                {"score": 0.6, "text": "sentence 1"},
                {"score": 0.5, "text": "sentence 2"},
            ],
        },
        "predictions": {
            "test_sentences_sorted_by_score": {
                "sentences": [
                    {"score": 1.0, "text": "sentence " "3"},
                    {"score": 0.6, "text": "sentence " "1"},
                    {"score": 0.5, "text": "sentence " "2"},
                ]
            }
        },
        "status": "Default",
        "text": "The input text",
    }


def test_model_with_predictions():
    record = Text2TextRecord.parse_obj(
        {
            "id": 0,
            "text": "The input text",
            "predictions": {
                "test_sentences_sorted_by_score": {
                    "sentences": [
                        {"score": 1.0, "text": "sentence " "3"},
                        {"score": 0.6, "text": "sentence " "1"},
                        {"score": 0.5, "text": "sentence " "2"},
                    ]
                }
            },
            "status": "Default",
        }
    )
    assert record.dict(exclude_none=True) == {
        "id": 0,
        "metrics": {},
        "prediction": {
            "agent": "test_sentences_sorted_by_score",
            "sentences": [
                {"score": 1.0, "text": "sentence 3"},
                {"score": 0.6, "text": "sentence 1"},
                {"score": 0.5, "text": "sentence 2"},
            ],
        },
        "predictions": {
            "test_sentences_sorted_by_score": {
                "sentences": [
                    {"score": 1.0, "text": "sentence " "3"},
                    {"score": 0.6, "text": "sentence " "1"},
                    {"score": 0.5, "text": "sentence " "2"},
                ]
            }
        },
        "status": "Default",
        "text": "The input text",
    }


def test_query_as_elasticsearch():
    query = Text2TextQuery(ids=[1, 2, 3])
    assert EsQueryBuilder._to_es_query(query) == {"ids": {"values": query.ids}}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
from typing import TYPE_CHECKING, AsyncGenerator, Generator

import httpx
import pytest
import pytest_asyncio
from argilla_server.cli.database.migrate import migrate_db
from argilla_server.database import database_url_sync
from argilla_server.settings import settings
from sqlalchemy import NullPool, create_engine
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine

from tests.database import SyncTestSession, TestSession, set_task

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection
    from sqlalchemy.ext.asyncio import AsyncConnection


# @pytest.fixture(scope="session")
# def event_loop() -> Generator["asyncio.AbstractEventLoop", None, None]:
#     loop = asyncio.get_event_loop_policy().get_event_loop()
#     yield loop
#     loop.close()


@pytest.fixture(scope="function")
def mock_httpx_client(mocker) -> Generator[httpx.Client, None, None]:
    return mocker.Mock(httpx.Client)


@pytest_asyncio.fixture(scope="session")
async def connection() -> AsyncGenerator["AsyncConnection", None]:
    set_task(asyncio.current_task())
    database_url = settings.database_url
    engine = create_async_engine(database_url, poolclass=NullPool)
    conn = await engine.connect()
    TestSession.configure(bind=conn)
    migrate_db("head")

    yield conn

    migrate_db("base")
    await conn.close()
    await engine.dispose()


@pytest_asyncio.fixture(autouse=True)
async def db(connection: "AsyncConnection") -> AsyncGenerator["AsyncSession", None]:
    await connection.begin_nested()
    session = TestSession()

    yield session

    await session.close()
    await TestSession.remove()
    await connection.rollback()


@pytest.fixture(scope="session")
def sync_connection() -> Generator["Connection", None, None]:
    engine = create_engine(database_url_sync())
    conn = engine.connect()
    SyncTestSession.configure(bind=conn)
    migrate_db("head")

    yield conn

    migrate_db("base")
    conn.close()
    engine.dispose()


@pytest.fixture(autouse=True)
def sync_db(sync_connection: "Connection") -> Generator["Session", None, None]:
    sync_connection.begin_nested()
    session = SyncTestSession()

    yield session

    session.close()
    SyncTestSession.remove()
    sync_connection.rollback()


@pytest.fixture
def async_db_proxy(mocker: "MockerFixture", sync_db: "Session") -> "AsyncSession":
    """Create a mocked `AsyncSession` that proxies to the sync session. This will allow us to execute the async CLI commands
    and then in the unit test function use the sync session to assert the changes.

    Args:
        mocker: pytest-mock fixture.
        sync_db: Sync session.

    Returns:
        Mocked `AsyncSession` that proxies to the sync session.
    """
    async_session = AsyncSession()
    async_session.sync_session = sync_db
    async_session._proxied = sync_db
    async_session.close = mocker.AsyncMock()

    return async_session



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Security, status
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import datasets
from argilla_server.database import get_async_db
from argilla_server.models import User
from argilla_server.policies import FieldPolicyV1, authorize
from argilla_server.schemas.v1.fields import Field as FieldSchema
from argilla_server.schemas.v1.fields import FieldUpdate
from argilla_server.security import auth

if TYPE_CHECKING:
    from argilla_server.models import Field

router = APIRouter(tags=["fields"])


async def _get_field(db: "AsyncSession", field_id: UUID) -> "Field":
    field = await datasets.get_field_by_id(db, field_id)
    if not field:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Field with id `{field_id}` not found",
        )
    return field


@router.patch("/fields/{field_id}", response_model=FieldSchema)
async def update_field(
    *,
    db: AsyncSession = Depends(get_async_db),
    field_id: UUID,
    field_update: FieldUpdate,
    current_user: User = Security(auth.get_current_user),
):
    field = await _get_field(db, field_id)

    await authorize(current_user, FieldPolicyV1.update(field))

    if field_update.settings and field_update.settings.type != field.settings["type"]:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Field type cannot be changed. Expected '{field.settings['type']}' but got '{field_update.settings.type}'",
        )

    return await datasets.update_field(db, field, field_update)


@router.delete("/fields/{field_id}", response_model=FieldSchema)
async def delete_field(
    *, db: AsyncSession = Depends(get_async_db), field_id: UUID, current_user: User = Security(auth.get_current_user)
):
    field = await _get_field(db, field_id)

    await authorize(current_user, FieldPolicyV1.delete(field))

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        await datasets.delete_field(db, field)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))

    return field



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Callable

import pytest
from argilla_server.apis.v0.models.commons.model import BulkResponse, SortableField
from argilla_server.apis.v0.models.token_classification import (
    TokenClassificationBulkRequest,
    TokenClassificationQuery,
    TokenClassificationRecord,
    TokenClassificationSearchRequest,
    TokenClassificationSearchResults,
)
from argilla_server.commons.models import TaskType
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User


@pytest.mark.asyncio
async def test_load_as_different_task(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_load_as_different_task"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "workspace": argilla_user.username, "task": TaskType.token_classification.value},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    expected_text = "This is a text with !"
    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [{"tokens": expected_text.split(" "), "raw_text": expected_text}]
    ]
    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TextClassification:search", json={}, params=workspace_query_params
    )
    assert response.status_code == 400

    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::WrongTaskError",
            "params": {"message": "Provided task TextClassification cannot be " "applied to dataset"},
        }
    }


@pytest.mark.asyncio
async def test_search_special_characters(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_search_special_characters"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "workspace": argilla_user.username, "task": TaskType.token_classification.value},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    expected_text = "This is a text with !"
    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [{"tokens": expected_text.split(" "), "raw_text": expected_text}]
    ]
    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:search",
        json=TokenClassificationSearchRequest(query=TokenClassificationQuery(query_text="\!")).dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert results.total == 0

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:search",
        json=TokenClassificationSearchRequest(query=TokenClassificationQuery(query_text="text.exact:\!")).dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert results.total == 1


@pytest.mark.asyncio
async def test_some_sort(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_some_sort"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "workspace": argilla_user.username, "task": TaskType.token_classification.value},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    expected_text = "This is a text with !"
    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [{"tokens": expected_text.split(" "), "raw_text": expected_text}]
    ]
    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:search",
        json=TokenClassificationSearchRequest(sort=[SortableField(id="babba")]).dict(),
        params=workspace_query_params,
    )
    assert response.status_code == 400

    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::BadRequestError",
            "params": {
                "message": "Wrong sort id babba. Valid values are: "
                "['id', 'metadata', 'score', 'predicted', "
                "'predicted_as', 'predicted_by', "
                "'annotated_as', 'annotated_by', 'status', "
                "'last_updated', 'event_timestamp']"
            },
        }
    }


@pytest.mark.parametrize(
    ("include_metrics", "metrics_validator"),
    [
        (True, lambda r: len(r.metrics) > 0),
        (False, lambda r: len(r.metrics) == 0),
        (None, lambda r: len(r.metrics) == 0),
    ],
)
@pytest.mark.asyncio
async def test_create_records_for_token_classification(
    async_client: "AsyncClient", argilla_user: User, include_metrics: bool, metrics_validator: Callable
):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_create_records_for_token_classification"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "workspace": argilla_user.username, "task": TaskType.token_classification.value},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    entity_label = "TEST"
    expected_records = 2
    record = {
        "tokens": "This is a text".split(" "),
        "raw_text": "This is a text",
        "metadata": {"field_one": "value one", "field_two": "value 2"},
        "prediction": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_label}]},
        "annotation": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_label}]},
    }
    records = [TokenClassificationRecord.parse_obj(record)] * expected_records

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == expected_records

    params = {**workspace_query_params}
    if include_metrics is not None:
        params["include_metrics"] = include_metrics

    response = await async_client.post(f"/api/datasets/{dataset}/TokenClassification:search", params=params)
    assert response.status_code == 200, response.json()

    results = TokenClassificationSearchResults.parse_obj(response.json())

    assert results.aggregations.dict(exclude={"score"}) == {
        "annotated_as": {"TEST": 1},
        "annotated_by": {"test": 1},
        "mentions": {"TEST": {"This": 1}},
        "metadata": {"field_one": {"value one": 1}, "field_two": {"value 2": 1}},
        "predicted": {"ok": 1},
        "predicted_as": {"TEST": 1},
        "predicted_by": {"test": 1},
        "predicted_mentions": {"TEST": {"This": 1}},
        "status": {"Default": 1},
        "words": {"text": 1},
    }

    assert "This" in results.aggregations.predicted_mentions[entity_label]
    assert "This" in results.aggregations.mentions[entity_label]
    for record in results.records:
        assert metrics_validator(record)


@pytest.mark.parametrize(
    ("include_metrics", "metrics_validator"),
    [
        (True, lambda r: len(r.metrics) > 0),
        (False, lambda r: len(r.metrics) == 0),
        (None, lambda r: len(r.metrics) == 0),
    ],
)
@pytest.mark.asyncio
async def test_create_records_for_token_classification_vector_search(
    async_client: "AsyncClient", argilla_user: User, include_metrics: bool, metrics_validator: Callable
):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_create_records_for_token_classification_vector_search"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "workspace": argilla_user.username, "task": TaskType.token_classification.value},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    entity_label = "TEST"
    expected_records = 3
    record_dicts = [
        {
            "id": 0,
            "tokens": "This is a text".split(" "),
            "raw_text": "This is a text",
            "metadata": {"field_one": "value one", "field_two": "value 2"},
            "prediction": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_label}]},
            "annotation": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_label}]},
            "vectors": {"my_bert": {"value": [1, 2, 3, 4, 5, 6]}},
        },
        {
            "id": 1,
            "tokens": "This is a text".split(" "),
            "raw_text": "This is a text",
            "metadata": {"field_one": "value one", "field_two": "value 2"},
            "prediction": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_label}]},
            "annotation": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_label}]},
            "vectors": {"my_bert": {"value": [5, 6, 7, 8, 9, 10]}},
        },
        {
            "id": 2,
            "tokens": "This is a text".split(" "),
            "raw_text": "This is a text",
            "metadata": {"field_one": "value one", "field_two": "value 2"},
            "prediction": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_label}]},
            "annotation": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_label}]},
            "vectors": {"my_bert": {"value": [7, 8, 9, 10, 11, 12]}},
        },
    ]

    records = [TokenClassificationRecord.parse_obj(record) for record in record_dicts]

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )

    assert response.status_code == 200, response.json()

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == expected_records

    params = {**workspace_query_params}
    if include_metrics is not None:
        params["include_metrics"] = include_metrics

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:search",
        json={"query": {"vector": {"name": "my_bert", "value": [5, 6, 7, 8, 9, 10]}}},
        params=params,
    )
    assert response.status_code == 200, response.json()

    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert results.aggregations.dict(exclude={"score"}) == {
        "annotated_as": {"TEST": 3},
        "annotated_by": {"test": 3},
        "mentions": {"TEST": {"This": 3}},
        "metadata": {"field_one": {"value one": 3}, "field_two": {"value 2": 3}},
        "predicted": {"ok": 3},
        "predicted_as": {"TEST": 3},
        "predicted_by": {"test": 3},
        "predicted_mentions": {"TEST": {"This": 3}},
        "status": {"Default": 3},
        "words": {"text": 3},
    }

    assert "This" in results.aggregations.predicted_mentions[entity_label]
    assert "This" in results.aggregations.mentions[entity_label]

    expected_record_ids = [1, 2, 0]
    for index, record in enumerate(results.records):
        assert metrics_validator(record)
        assert expected_record_ids[index] == record.id
        assert hasattr(record, "vectors")


@pytest.mark.asyncio
async def test_multiple_mentions_in_same_record(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_multiple_mentions_in_same_record"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "workspace": argilla_user.username, "task": TaskType.token_classification.value},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    entity_a, entity_b = ("TESTA", "TESTB")
    records = [
        TokenClassificationRecord.parse_obj(data)
        for data in [
            {
                "tokens": "This is a text".split(" "),
                "raw_text": "This is a text",
                "metadata": {"field_one": "value one", "field_two": "value 2"},
                "prediction": {
                    "agent": "test",
                    "entities": [{"start": 0, "end": 4, "label": entity_a}, {"start": 5, "end": 7, "label": entity_b}],
                },
            },
            {
                "tokens": "This is a text".split(" "),
                "raw_text": "This is a text",
                "metadata": {"field_one": "value one", "field_two": "value 2"},
                "prediction": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_a}]},
            },
            {
                "tokens": "This is a text".split(" "),
                "raw_text": "This is a text",
                "metadata": {"field_one": "value one", "field_two": "value 2"},
                "annotation": {"agent": "test", "entities": [{"start": 0, "end": 4, "label": entity_a}]},
            },
        ]
    ]
    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkRequest(
            tags={"env": "test", "class": "text classification"},
            metadata={"config": {"the": "config"}},
            records=records,
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    bulk_response = BulkResponse.parse_obj(response.json())
    assert bulk_response.dataset == dataset
    assert bulk_response.failed == 0
    assert bulk_response.processed == 3

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:search", params=workspace_query_params
    )
    assert response.status_code == 200, response.json()

    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert "This" in results.aggregations.predicted_mentions[entity_a]
    assert results.aggregations.predicted_mentions[entity_a]["This"] == 2
    assert "is" in results.aggregations.predicted_mentions[entity_b]
    assert results.aggregations.predicted_mentions[entity_b]["is"] == 1


@pytest.mark.asyncio
async def test_show_not_aggregable_metadata_fields(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = "test_show_not_aggregable_metadata_fields"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "workspace": argilla_user.username, "task": TaskType.token_classification.value},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkRequest(
            records=[
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 1.3, "field_two": 2},
                    }
                ),
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 2.3, "field_two": 300},
                    },
                ),
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 11.333, "field_two": -10},
                    },
                ),
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:search", json={}, params=workspace_query_params
    )
    assert response.status_code == 200, response.json()

    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert len(results.aggregations.metadata) == 2
    assert "field_one" in results.aggregations.metadata
    assert "argilla:stats" in results.aggregations.metadata["field_one"]


@pytest.mark.asyncio
async def test_search_with_raw_query(async_client: "AsyncClient", argilla_user: User):
    async_client.headers.update({API_KEY_HEADER_NAME: argilla_user.api_key})
    workspace_query_params = {"workspace": argilla_user.username}

    dataset = await log_some_data_for_token_classification(async_client, workspace_name=argilla_user.username)

    raw_es_query_matching_all = {
        "bool": {
            "filter": {
                "bool": {
                    "must": {"script": {"script": {"source": "doc['tokens'].value.length() == 4", "lang": "painless"}}}
                }
            }
        }
    }

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:search",
        json={"query": {"raw_query": raw_es_query_matching_all}},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    results = TokenClassificationSearchResults.parse_obj(response.json())
    assert len(results.records) == 3


async def log_some_data_for_token_classification(async_client: "AsyncClient", workspace_name: str):
    workspace_query_params = {"workspace": workspace_name}

    dataset = "log_some_data_for_token_classification"
    response = await async_client.delete(f"/api/datasets/{dataset}", params=workspace_query_params)
    assert response.status_code == 200

    response = await async_client.post(
        "/api/datasets",
        json={"name": dataset, "workspace": "argilla", "task": TaskType.token_classification.value},
        params=workspace_query_params,
    )
    assert response.status_code == 200

    response = await async_client.post(
        f"/api/datasets/{dataset}/TokenClassification:bulk",
        json=TokenClassificationBulkRequest(
            records=[
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 1.3, "field_two": 2},
                    },
                ),
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 2.3, "field_two": 300},
                    },
                ),
                TokenClassificationRecord.parse_obj(
                    {
                        "tokens": "This is a text".split(" "),
                        "raw_text": "This is a text",
                        "metadata": {"field_one": 11.333, "field_two": -10},
                    },
                ),
            ],
        ).dict(by_alias=True),
        params=workspace_query_params,
    )
    assert response.status_code == 200, response.json()

    return dataset



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import TYPE_CHECKING
from uuid import uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import DatasetStatus, Field, UserRole
from argilla_server.schemas.v1.fields import FIELD_CREATE_TITLE_MAX_LENGTH
from sqlalchemy import func, select

from tests.factories import (
    AnnotatorFactory,
    DatasetFactory,
    TextFieldFactory,
    UserFactory,
)

if TYPE_CHECKING:
    from httpx import AsyncClient
    from sqlalchemy.ext.asyncio import AsyncSession


@pytest.mark.parametrize(
    "payload, expected_settings",
    [
        (
            {"title": "New Title", "settings": {"type": "text", "use_markdown": True}},
            {"type": "text", "use_markdown": True},
        ),
        ({"title": "New Title"}, {"type": "text", "use_markdown": False}),
        (
            {"name": "New Name", "required": True, "dataset_id": str(uuid4())},
            {"type": "text", "use_markdown": False},
        ),
    ],
)
@pytest.mark.parametrize("role", [UserRole.admin, UserRole.owner])
@pytest.mark.asyncio
async def test_update_field(
    async_client: "AsyncClient", db: "AsyncSession", role: UserRole, payload: dict, expected_settings: dict
):
    field = await TextFieldFactory.create()
    user = await UserFactory.create(role=role, workspaces=[field.dataset.workspace])

    response = await async_client.patch(
        f"/api/v1/fields/{field.id}", headers={API_KEY_HEADER_NAME: user.api_key}, json=payload
    )

    title = payload.get("title") or field.title

    assert response.status_code == 200
    assert response.json() == {
        "id": str(field.id),
        "name": field.name,
        "title": title,
        "required": field.required,
        "settings": expected_settings,
        "dataset_id": str(field.dataset.id),
        "inserted_at": field.inserted_at.isoformat(),
        "updated_at": field.updated_at.isoformat(),
    }

    field = await db.get(Field, field.id)
    assert field.title == title
    assert field.settings == expected_settings


@pytest.mark.parametrize("title", [None, "", "t" * (FIELD_CREATE_TITLE_MAX_LENGTH + 1)])
@pytest.mark.asyncio
async def test_update_field_with_invalid_title(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict, title: str
):
    field = await TextFieldFactory.create(title="title")

    response = await async_client.patch(f"/api/v1/fields/{field.id}", headers=owner_auth_header, json={"title": title})

    assert response.status_code == 422

    field = await db.get(Field, field.id)
    assert field.title == "title"


@pytest.mark.parametrize(
    "payload",
    [
        {"settings": None},
        {"settings": {"type": "text"}},
        {"settings": {"type": "text", "use_markdown": None}},
        {"settings": {"type": "i don't exist"}},
    ],
)
@pytest.mark.asyncio
async def test_update_field_with_invalid_settings(async_client: "AsyncClient", owner_auth_header: dict, payload: dict):
    field = await TextFieldFactory.create()

    response = await async_client.patch(f"/api/v1/fields/{field.id}", headers=owner_auth_header, json=payload)

    assert response.status_code == 422


@pytest.mark.asyncio
async def test_update_field_with_invalid_payload(async_client: "AsyncClient", owner_auth_header: dict):
    field = await TextFieldFactory.create()

    response = await async_client.patch(
        f"/api/v1/fields/{field.id}",
        headers=owner_auth_header,
        json={"title": {"this": "is", "not": "valid"}, "settings": {"use_markdown": "no"}},
    )

    assert response.status_code == 422


@pytest.mark.asyncio
async def test_update_field_non_existent(async_client: "AsyncClient", owner_auth_header: dict):
    response = await async_client.patch(
        f"/api/v1/fields/{uuid4()}",
        headers=owner_auth_header,
        json={"title": "New Title", "settings": {"type": "text", "use_markdown": True}},
    )

    assert response.status_code == 404


@pytest.mark.asyncio
async def test_update_field_as_admin_from_different_workspace(async_client: "AsyncClient"):
    field = await TextFieldFactory.create()
    user = await UserFactory.create(role=UserRole.admin)

    response = await async_client.patch(
        f"/api/v1/fields/{field.id}",
        headers={API_KEY_HEADER_NAME: user.api_key},
        json={"title": "New Title", "settings": {"type": "text", "use_markdown": True}},
    )

    assert response.status_code == 403


@pytest.mark.asyncio
async def test_update_field_as_annotator(async_client: "AsyncClient"):
    field = await TextFieldFactory.create()
    user = await UserFactory.create(role=UserRole.annotator, workspaces=[field.dataset.workspace])

    response = await async_client.patch(
        f"/api/v1/fields/{field.id}",
        headers={API_KEY_HEADER_NAME: user.api_key},
        json={"title": "New Title", "settings": {"type": "text", "use_markdown": True}},
    )

    assert response.status_code == 403


@pytest.mark.asyncio
async def test_update_field_without_authentication(async_client: "AsyncClient"):
    field = await TextFieldFactory.create()

    response = await async_client.patch(
        f"/api/v1/fields/{field.id}",
        json={"title": "New Title", "settings": {"use_markdown": True}},
    )

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_delete_field(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    field = await TextFieldFactory.create(name="name", title="title")

    response = await async_client.delete(f"/api/v1/fields/{field.id}", headers=owner_auth_header)

    assert response.status_code == 200
    assert (await db.execute(select(func.count(Field.id)))).scalar() == 0

    response_body = response.json()
    assert response_body == {
        "id": str(field.id),
        "name": "name",
        "title": "title",
        "required": False,
        "settings": {"type": "text", "use_markdown": False},
        "dataset_id": str(field.dataset.id),
        "inserted_at": datetime.fromisoformat(response_body["inserted_at"]).isoformat(),
        "updated_at": datetime.fromisoformat(response_body["updated_at"]).isoformat(),
    }


@pytest.mark.asyncio
async def test_delete_field_without_authentication(async_client: "AsyncClient", db: "AsyncSession"):
    field = await TextFieldFactory.create()

    response = await async_client.delete(f"/api/v1/fields/{field.id}")

    assert response.status_code == 401
    assert (await db.execute(select(func.count(Field.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_field_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()
    field = await TextFieldFactory.create()

    response = await async_client.delete(f"/api/v1/fields/{field.id}", headers={API_KEY_HEADER_NAME: annotator.api_key})

    assert response.status_code == 403
    assert (await db.execute(select(func.count(Field.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_field_belonging_to_published_dataset(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    dataset = await DatasetFactory.create(status=DatasetStatus.ready)
    field = await TextFieldFactory.create(dataset=dataset)

    response = await async_client.delete(f"/api/v1/fields/{field.id}", headers=owner_auth_header)

    assert response.status_code == 422
    assert response.json() == {"detail": "Fields cannot be deleted for a published dataset"}
    assert (await db.execute(select(func.count(Field.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_field_with_nonexistent_field_id(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    await TextFieldFactory.create()

    response = await async_client.delete(f"/api/v1/fields/{uuid4()}", headers=owner_auth_header)

    assert response.status_code == 404
    assert (await db.execute(select(func.count(Field.id)))).scalar() == 1



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.commons.models import TaskType
from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.daos.datasets import DatasetsDAO
from argilla_server.daos.models.datasets import BaseDatasetDB
from argilla_server.daos.records import DatasetRecordsDAO
from argilla_server.errors import ClosedDatasetError

es_wrapper = GenericElasticEngineBackend.get_instance()
records = DatasetRecordsDAO.get_instance(es_wrapper)
dao = DatasetsDAO.get_instance(es_wrapper, records)


def test_retrieve_ownered_dataset_for_no_owner_user():
    dataset = "test_retrieve_owned_dataset_for_no_owner_user"
    created = dao.create_dataset(
        BaseDatasetDB(name=dataset, workspace="other", task=TaskType.text_classification),
    )
    assert dao.find_by_name(created.name, workspace=created.workspace) == created
    assert dao.find_by_name(created.name, workspace="me") is None


def test_list_datasets_by_task():
    dataset = "test_list_datasets_by_task"
    workspace_name = "other"

    all_datasets = dao.list_datasets(workspaces=[workspace_name])
    for ds in all_datasets:
        dao.delete_dataset(ds)

    created_text = dao.create_dataset(
        BaseDatasetDB(
            name=dataset + "_text",
            workspace=workspace_name,
            task=TaskType.text_classification,
        ),
    )

    created_token = dao.create_dataset(
        BaseDatasetDB(
            name=dataset + "_token",
            workspace=workspace_name,
            task=TaskType.token_classification,
        ),
    )

    assert len(dao.list_datasets()) == 0
    assert len(dao.list_datasets(workspaces=[workspace_name])) == 2

    datasets = dao.list_datasets(workspaces=[workspace_name], task2dataset_map={created_text.task: BaseDatasetDB})

    assert len(datasets) == 1
    assert datasets[0].name == created_text.name

    datasets = dao.list_datasets(workspaces=[workspace_name], task2dataset_map={created_token.task: BaseDatasetDB})

    assert len(datasets) == 1
    assert datasets[0].name == created_token.name


def test_close_dataset():
    dataset = "test_close_dataset"

    created = dao.create_dataset(
        BaseDatasetDB(name=dataset, workspace="other", task=TaskType.text_classification),
    )

    dao.close(created)
    with pytest.raises(ClosedDatasetError, match=dataset):
        records.search_records(dataset=created)

    dao.open(created)
    records.search_records(dataset=created)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Dict, List, Optional

from argilla_server.apis.v0.models.commons.model import BaseRecord, BaseRecordInputs, BaseSearchResults, ScoreRange
from argilla_server.commons.models import PredictionStatus
from argilla_server.daos.backend.search.model import SortableField
from argilla_server.pydantic_v1 import BaseModel, Field, root_validator, validator
from argilla_server.schemas.v0.datasets import UpdateDatasetRequest
from argilla_server.services.search.model import ServiceBaseRecordsQuery, ServiceBaseSearchResultsAggregations
from argilla_server.services.tasks.token_classification.model import (
    ServiceTokenClassificationAnnotation as _TokenClassificationAnnotation,
)


class TokenClassificationAnnotation(_TokenClassificationAnnotation):
    pass


class TokenClassificationRecordInputs(BaseRecordInputs[TokenClassificationAnnotation]):
    text: str = Field()
    tokens: List[str] = Field(min_items=1)
    # TODO(@frascuchon): Delete this field and all related logic
    _raw_text: Optional[str] = Field(alias="raw_text")

    @root_validator(pre=True)
    def accept_old_fashion_text_field(cls, values):
        text, raw_text = values.get("text"), values.get("raw_text")
        text = text or raw_text
        values["text"] = cls.check_text_content(text)

        return values

    @validator("text")
    def check_text_content(cls, text: str):
        assert text and text.strip(), "No text or empty text provided"
        return text


class TokenClassificationRecord(TokenClassificationRecordInputs, BaseRecord[TokenClassificationAnnotation]):
    pass


class TokenClassificationBulkRequest(UpdateDatasetRequest):
    records: List[TokenClassificationRecordInputs]


class TokenClassificationQuery(ServiceBaseRecordsQuery):
    predicted_as: List[str] = Field(default_factory=list)
    annotated_as: List[str] = Field(default_factory=list)
    score: Optional[ScoreRange] = Field(default=None)
    predicted: Optional[PredictionStatus] = Field(default=None, nullable=True)


class TokenClassificationSearchRequest(BaseModel):
    query: TokenClassificationQuery = Field(default_factory=TokenClassificationQuery)
    sort: List[SortableField] = Field(default_factory=list)


class TokenClassificationAggregations(ServiceBaseSearchResultsAggregations):
    predicted_mentions: Dict[str, Dict[str, int]] = Field(default_factory=dict)
    mentions: Dict[str, Dict[str, int]] = Field(default_factory=dict)


class TokenClassificationSearchResults(BaseSearchResults[TokenClassificationRecord, TokenClassificationAggregations]):
    pass



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from typing import TYPE_CHECKING

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import UserRole

from tests.factories import UserFactory, WorkspaceFactory

if TYPE_CHECKING:
    from fastapi.testclient import TestClient


@pytest.mark.asyncio
class TestsUsersV1Endpoints:
    async def test_list_user_workspaces(self, async_client: "AsyncClient", owner_auth_header: dict):
        workspaces = await WorkspaceFactory.create_batch(3)
        user = await UserFactory.create(workspaces=workspaces)

        response = await async_client.get(f"/api/v1/users/{user.id}/workspaces", headers=owner_auth_header)

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "id": str(workspace.id),
                    "name": workspace.name,
                    "inserted_at": workspace.inserted_at.isoformat(),
                    "updated_at": workspace.updated_at.isoformat(),
                }
                for workspace in workspaces
            ]
        }

    async def test_list_user_workspaces_for_owner(self, async_client: "AsyncClient"):
        workspaces = await WorkspaceFactory.create_batch(5)
        owner = await UserFactory.create(role=UserRole.owner)

        response = await async_client.get(
            f"/api/v1/users/{owner.id}/workspaces", headers={API_KEY_HEADER_NAME: owner.api_key}
        )
        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "id": str(workspace.id),
                    "name": workspace.name,
                    "inserted_at": workspace.inserted_at.isoformat(),
                    "updated_at": workspace.updated_at.isoformat(),
                }
                for workspace in workspaces
            ]
        }

    @pytest.mark.parametrize("role", [UserRole.annotator, UserRole.admin])
    async def test_list_user_workspaces_as_restricted_user(self, async_client: "AsyncClient", role: UserRole):
        workspaces = await WorkspaceFactory.create_batch(3)
        user = await UserFactory.create(workspaces=workspaces)
        requesting_user = await UserFactory.create(role=role)

        response = await async_client.get(
            f"/api/v1/users/{user.id}/workspaces", headers={API_KEY_HEADER_NAME: requesting_user.api_key}
        )

        assert response.status_code == 403

    async def test_list_user_for_non_existing_user(self, async_client: "AsyncClient", owner_auth_header: dict):
        response = await async_client.get(f"/api/v1/users/{uuid.uuid4()}/workspaces", headers=owner_auth_header)
        assert response.status_code == 404



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from fastapi import APIRouter, Depends, Query, Security

from argilla_server.apis.v0.handlers import metrics, token_classification_dataset_settings
from argilla_server.apis.v0.models.commons.model import BulkResponse
from argilla_server.apis.v0.models.commons.params import CommonTaskHandlerDependencies, RequestPagination
from argilla_server.apis.v0.models.token_classification import (
    TokenClassificationAggregations,
    TokenClassificationBulkRequest,
    TokenClassificationQuery,
    TokenClassificationRecord,
    TokenClassificationSearchRequest,
    TokenClassificationSearchResults,
)
from argilla_server.apis.v0.validators.token_classification import DatasetValidator
from argilla_server.commons.config import TasksFactory
from argilla_server.commons.models import TaskType
from argilla_server.models import User
from argilla_server.security import auth
from argilla_server.services.datasets import DatasetsService
from argilla_server.services.tasks.token_classification import TokenClassificationService
from argilla_server.services.tasks.token_classification.metrics import TokenClassificationMetrics
from argilla_server.services.tasks.token_classification.model import (
    ServiceTokenClassificationQuery,
    ServiceTokenClassificationRecord,
)


def configure_router():
    task_type = TaskType.token_classification
    base_endpoint = f"/{{name}}/{task_type.value}"

    TasksFactory.register_task(
        task_type=task_type,
        query_request=TokenClassificationQuery,
        record_class=ServiceTokenClassificationRecord,
        metrics=TokenClassificationMetrics,
    )

    router = APIRouter(tags=[task_type], prefix="/datasets")

    @router.post(
        path=f"{base_endpoint}:bulk",
        operation_id="bulk_records",
        response_model=BulkResponse,
        response_model_exclude_none=True,
    )
    async def bulk_records(
        name: str,
        bulk: TokenClassificationBulkRequest,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TokenClassificationService = Depends(TokenClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        validator: DatasetValidator = Depends(DatasetValidator.get_instance),
        current_user: User = Security(auth.get_current_user, scopes=[]),
    ) -> BulkResponse:
        task = task_type
        workspace = common_params.workspace

        dataset = await datasets.find_by_name(current_user, name=name, task=task, workspace=workspace)

        await datasets.update(user=current_user, dataset=dataset, tags=bulk.tags, metadata=bulk.metadata)

        records = [ServiceTokenClassificationRecord.parse_obj(r) for r in bulk.records]
        # TODO(@frascuchon): validator can be applied in service layer
        await validator.validate_dataset_records(user=current_user, dataset=dataset, records=records)

        result = await service.add_records(dataset=dataset, records=records)
        return BulkResponse(dataset=name, processed=result.processed, failed=result.failed)

    @router.post(
        path=f"{base_endpoint}:search",
        response_model=TokenClassificationSearchResults,
        response_model_exclude_none=True,
        operation_id="search_records",
    )
    async def search_records(
        name: str,
        search: TokenClassificationSearchRequest = None,
        common_params: CommonTaskHandlerDependencies = Depends(),
        include_metrics: bool = Query(
            default=False,
            description="If enabled, return related record metrics",
        ),
        pagination: RequestPagination = Depends(),
        service: TokenClassificationService = Depends(TokenClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user, scopes=[]),
    ) -> TokenClassificationSearchResults:
        search = search or TokenClassificationSearchRequest()
        query = search.query or TokenClassificationQuery()

        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
        )
        results = service.search(
            dataset=dataset,
            query=ServiceTokenClassificationQuery.parse_obj(query),
            sort_by=search.sort,
            record_from=pagination.from_,
            size=pagination.limit,
            exclude_metrics=not include_metrics,
        )

        return TokenClassificationSearchResults(
            total=results.total,
            records=[TokenClassificationRecord.parse_obj(r) for r in results.records],
            aggregations=TokenClassificationAggregations.parse_obj(results.metrics) if results.metrics else None,
        )

    token_classification_dataset_settings.configure_router(router)
    metrics.configure_router(
        router,
        cfg=TasksFactory.get_task_by_task_type(task_type),
    )

    return router


router = configure_router()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""create vectors settings table

Revision ID: 7850ab5b42d9
Revises: 84f6b9ff6076
Create Date: 2023-09-12 16:21:14.321044

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7850ab5b42d9"
down_revision = "84f6b9ff6076"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "vectors_settings",
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("title", sa.Text(), nullable=False),
        sa.Column("dimensions", sa.Integer(), nullable=False),
        sa.Column("dataset_id", sa.Uuid(), nullable=False),
        sa.Column("id", sa.Uuid(), nullable=False),
        sa.Column("inserted_at", sa.DateTime(), nullable=False),
        sa.Column("updated_at", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["dataset_id"], ["datasets.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name", "dataset_id", name="vector_settings_name_dataset_id_uq"),
    )
    op.create_index(op.f("ix_vectors_settings_dataset_id"), "vectors_settings", ["dataset_id"], unique=False)
    op.create_index(op.f("ix_vectors_settings_name"), "vectors_settings", ["name"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_vectors_settings_name"), table_name="vectors_settings")
    op.drop_index(op.f("ix_vectors_settings_dataset_id"), table_name="vectors_settings")
    op.drop_table("vectors_settings")
    # ### end Alembic commands ###



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import alembic.config
import typer

from argilla_server.database import ALEMBIC_CONFIG_FILE, TAGGED_REVISIONS

from . import utils


def revisions():
    current_revision = utils.get_current_revision(ALEMBIC_CONFIG_FILE, verbose=True)

    typer.echo("")
    typer.echo("Tagged revisions")
    typer.echo("-----------------")
    for version, revision in TAGGED_REVISIONS.items():
        typer.echo(f"• {version} (revision: {revision!r})")

    typer.echo("")
    typer.echo("Alembic revisions")
    typer.echo("-----------------")
    alembic.config.main(argv=["-c", ALEMBIC_CONFIG_FILE, "history"])

    typer.echo("")
    typer.echo("Current revision")
    typer.echo("----------------")
    typer.echo(current_revision)


if __name__ == "__main__":
    typer.run(revisions)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import pytest
from argilla_server.utils.span_utils import SpanUtils


def test_init():
    text = "test this."
    tokens = ["test", "this", "."]

    span_utils = SpanUtils(text, tokens)

    assert span_utils.text is text
    assert span_utils.tokens is tokens

    assert span_utils.token_to_char_idx == {0: (0, 4), 1: (5, 9), 2: (9, 10)}
    assert span_utils.char_to_token_idx == {
        0: 0,
        1: 0,
        2: 0,
        3: 0,
        5: 1,
        6: 1,
        7: 1,
        8: 1,
        9: 2,
    }

    assert span_utils._start_to_token_idx == {0: 0, 5: 1, 9: 2}
    assert span_utils._end_to_token_idx == {4: 0, 9: 1, 10: 2}


def test_init_value_error():
    with pytest.raises(ValueError, match="Token 'ValueError' not found in text: test error"):
        SpanUtils(text="test error", tokens=["test", "ValueError"])


def test_validate():
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    assert span_utils.validate([("mock", 5, 10)]) is None


def test_validate_not_valid_spans():
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    with pytest.raises(ValueError, match="Following entity spans are not valid: \[\('mock', 2, 1\)\]\n"):
        span_utils.validate([("mock", 2, 1)])


def test_validate_misaligned_spans():
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    with pytest.raises(
        ValueError,
        match="Following entity spans are not aligned with provided tokenization\n"
        r"Spans:\n\('mock', 0, 5\) - 'test '\n"
        r"Tokens:\n\['test', 'this', '.'\]",
    ):
        span_utils.validate([("mock", 0, 5)])


def test_validate_not_valid_and_misaligned_spans():
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    with pytest.raises(
        ValueError,
        match=r"Following entity spans are not valid: \[\('mock', 2, 1\)\]\n"
        "Following entity spans are not aligned with provided tokenization\n"
        r"Spans:\n\('mock', 0, 5\) - 'test '\n"
        r"Tokens:\n\['test', 'this', '.'\]",
    ):
        span_utils.validate([("mock", 2, 1), ("mock", 0, 5)])


@pytest.mark.parametrize(
    "spans, expected",
    [
        ([("mock", -1, 4), ("mock", 20, 22)], [("mock", 0, 4), ("mock", 20, 21)]),
        ([("mock", 0, 5), ("mock", 4, 9)], [("mock", 0, 4), ("mock", 5, 9)]),
        ([("mock", 10, 15), ("mock", 11, 16)], [("mock", 11, 15), ("mock", 11, 15)]),
    ],
)
def test_correct(spans, expected):
    text = "test this \nnext\ttext."
    tokens = ["test", "this", "\n", "next", "\t", "text", "."]
    span_utils = SpanUtils(text, tokens)

    assert span_utils.correct(spans) == expected


def test_to_tags_overlapping_spans():
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    with pytest.raises(ValueError, match="IOB tags cannot handle overlapping spans!"):
        span_utils.to_tags([("mock", 0, 4), ("mock", 0, 9)])


def test_to_tags():
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    tags = span_utils.to_tags([("mock", 0, 9)])
    assert tags == ["B-mock", "I-mock", "O"]


def test_from_tags_wrong_length():
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    with pytest.raises(
        ValueError,
        match="The list of tags must have the same length as the list of tokens!",
    ):
        span_utils.from_tags(["mock", "mock"])


def test_from_tags_not_valid_format():
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    with pytest.raises(ValueError, match="Tags are not in the IOB or BILOU format!"):
        span_utils.from_tags(["mock", "mock", "mock"])


@pytest.mark.parametrize(
    "tags,expected",
    [
        (["B-mock", "O", "O"], [("mock", 0, 4)]),
        (["I-mock", "O", "O"], [("mock", 0, 4)]),
        (["U-mock", "O", "O"], [("mock", 0, 4)]),
        (["L-mock", "O", "O"], [("mock", 0, 4)]),
        (["B-mock", "I-mock", "O"], [("mock", 0, 9)]),
        (["I-mock", "I-mock", "O"], [("mock", 0, 9)]),
        (["B-mock", "L-mock", "O"], [("mock", 0, 9)]),
        (["I-mock", "L-mock", "O"], [("mock", 0, 9)]),
        (["B-mock", "I-mock", "I-mock"], [("mock", 0, 10)]),
        (["I-mock", "I-mock", "I-mock"], [("mock", 0, 10)]),
        (["B-mock", "I-mock", "L-mock"], [("mock", 0, 10)]),
        (["B-mock", "L-mock", "L-mock"], [("mock", 0, 9), ("mock", 9, 10)]),
        (["U-mock", "U-mock", "O"], [("mock", 0, 4), ("mock", 5, 9)]),
        (["U-mock", "I-mock", "O"], [("mock", 0, 4), ("mock", 5, 9)]),
        (["B-mock", "B-mock", "O"], [("mock", 0, 4), ("mock", 5, 9)]),
        (["U-mock", "B-mock", "O"], [("mock", 0, 4), ("mock", 5, 9)]),
        (["I-mock", "B-mock", "O"], [("mock", 0, 4), ("mock", 5, 9)]),
        (["L-mock", "B-mock", "O"], [("mock", 0, 4), ("mock", 5, 9)]),
    ],
)
def test_from_tags(tags, expected):
    span_utils = SpanUtils("test this.", ["test", "this", "."])
    assert span_utils.from_tags(tags) == expected



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from typing import Any, Dict, Iterable, Optional, Tuple

import elasticsearch8

from argilla_server.daos.backend.base import BackendErrorHandler
from argilla_server.daos.backend.client_adapters.opensearch import OpenSearchClient
from argilla_server.daos.backend.search.query_builder import EsQueryBuilder

ES_CLIENT_VERSION: str = elasticsearch8.__versionstr__

from elasticsearch8 import ApiError, Elasticsearch, ElasticsearchWarning, NotFoundError, RequestError, helpers
from elasticsearch8.helpers import BulkIndexError


@dataclasses.dataclass
class ElasticsearchClient(OpenSearchClient):
    ES_CLIENT_VERSION = ES_CLIENT_VERSION
    query_builder = EsQueryBuilder()

    def __post_init__(self):
        self.__client__ = Elasticsearch(**self.config_backend)
        self.error_handling = BackendErrorHandler(
            WarningIgnore=ElasticsearchWarning,
            RequestError=RequestError,
            BulkError=BulkIndexError,
            NotFoundError=NotFoundError,
            GenericApiError=ApiError,
        )

    def configure_index_vectors(
        self,
        *,
        index: str,
        vectors: Dict[str, int],
        similarity_metric: str = "l2_norm",
    ):
        vector_mappings = {}
        for vector_name, vector_dimension in vectors.items():
            index_mapping = {
                "type": "dense_vector",
                "dims": vector_dimension,
                "index": True,
                # can similarity property also be part of config @frascuchon ?
                # relates vector search similarity metric
                "similarity": similarity_metric,  ## default value regarding the knn best practices es documentation
            }
            vector_field = self.query_builder.get_vector_field_name(vector_name)
            vector_mappings[vector_field] = index_mapping

        self.set_index_mappings(
            index=index,
            properties=vector_mappings,
        )

    def _reindex(
        self,
        *,
        source_index: str,
        target_index: str,
    ):
        with self.error_handling(index=source_index):
            helpers.reindex(
                client=self.__client__,
                source_index=source_index,
                target_index=target_index,
            )

    def bulk(
        self,
        *,
        index: str,
        actions: Iterable[dict],
    ) -> Tuple:
        with self.error_handling(index=index):
            return helpers.bulk(
                client=self.__client__,
                index=index,
                actions=actions,
                raise_on_error=True,
                refresh="wait_for",
            )

    def _es_search(
        self,
        index: str,
        es_query: Dict[str, Any],
        size: int,
        routing: Optional[str] = None,
    ):
        knn = es_query.pop("knn", None)
        if knn:
            results = self.__client__.search(
                index=index,
                knn=knn,
                query=es_query.get("query"),
                aggs=es_query.get("aggs"),
                routing=routing,
                track_total_hits=True,
                rest_total_hits_as_int=True,
                size=size,
                source=es_query.get("_source"),
            )
        else:
            results = self.__client__.search(
                index=index,
                routing=routing,
                body=es_query,
                track_total_hits=True,
                rest_total_hits_as_int=True,
                size=size,
            )
        return results



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional

from fastapi import APIRouter, Depends, Query, Security

from argilla_server.apis.v0.handlers import metrics, text_classification_dataset_settings
from argilla_server.apis.v0.helpers import deprecate_endpoint
from argilla_server.apis.v0.models.commons.model import BulkResponse
from argilla_server.apis.v0.models.commons.params import CommonTaskHandlerDependencies, RequestPagination
from argilla_server.apis.v0.models.text_classification import (
    CreateLabelingRule,
    DatasetLabelingRulesMetricsSummary,
    LabelingRule,
    LabelingRuleMetricsSummary,
    TextClassificationBulkRequest,
    TextClassificationDataset,
    TextClassificationQuery,
    TextClassificationSearchAggregations,
    TextClassificationSearchRequest,
    TextClassificationSearchResults,
    UpdateLabelingRule,
)
from argilla_server.apis.v0.validators.text_classification import DatasetValidator
from argilla_server.commons.config import TasksFactory
from argilla_server.commons.models import TaskType
from argilla_server.models import User
from argilla_server.security import auth
from argilla_server.services.datasets import DatasetsService
from argilla_server.services.tasks.text_classification import TextClassificationService
from argilla_server.services.tasks.text_classification.metrics import TextClassificationMetrics
from argilla_server.services.tasks.text_classification.model import (
    ServiceLabelingRule,
    ServiceTextClassificationQuery,
    ServiceTextClassificationRecord,
)


def configure_router():
    task_type = TaskType.text_classification
    base_endpoint = f"/{{name}}/{task_type.value}"
    new_base_endpoint = f"/{task_type.value}/{{name}}"

    TasksFactory.register_task(
        task_type=task_type,
        dataset_class=TextClassificationDataset,
        query_request=TextClassificationQuery,
        record_class=ServiceTextClassificationRecord,
        metrics=TextClassificationMetrics,
    )

    router = APIRouter(tags=[task_type], prefix="/datasets")

    @router.post(
        f"{base_endpoint}:bulk",
        operation_id="bulk_records",
        response_model=BulkResponse,
        response_model_exclude_none=True,
    )
    async def bulk_records(
        name: str,
        bulk: TextClassificationBulkRequest,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        validator: DatasetValidator = Depends(DatasetValidator.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> BulkResponse:
        task = task_type
        workspace = common_params.workspace

        dataset = await datasets.find_by_name(current_user, name=name, task=task, workspace=workspace)

        await datasets.update(user=current_user, dataset=dataset, tags=bulk.tags, metadata=bulk.metadata)

        # TODO(@frascuchon): Validator should be applied in the service layer
        records = [ServiceTextClassificationRecord.parse_obj(r) for r in bulk.records]
        await validator.validate_dataset_records(user=current_user, dataset=dataset, records=records)

        result = await service.add_records(dataset=dataset, records=records)
        return BulkResponse(dataset=name, processed=result.processed, failed=result.failed)

    @router.post(
        f"{base_endpoint}:search",
        response_model=TextClassificationSearchResults,
        response_model_exclude_none=True,
        operation_id="search_records",
    )
    async def search_records(
        name: str,
        search: TextClassificationSearchRequest = None,
        common_params: CommonTaskHandlerDependencies = Depends(),
        include_metrics: bool = Query(False, description="If enabled, return related record metrics"),
        pagination: RequestPagination = Depends(),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> TextClassificationSearchResults:
        """
        Searches data from dataset

        Parameters
        ----------
        name:
            The dataset name
        search:
            The search query request
        common_params:
            Common query params
        include_metrics:
            Flag to enable include metrics
        pagination:
            The pagination params
        service:
            The dataset records service
        datasets:
            The dataset service
        current_user:
            The current request user


        Returns
        -------
            The search results data

        """

        search = search or TextClassificationSearchRequest()
        query = search.query or TextClassificationQuery()
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
        )
        result = service.search(
            dataset=dataset,
            query=ServiceTextClassificationQuery.parse_obj(query),
            sort_by=search.sort,
            record_from=pagination.from_,
            size=pagination.limit,
            exclude_metrics=not include_metrics,
        )

        return TextClassificationSearchResults(
            total=result.total,
            records=result.records,
            aggregations=TextClassificationSearchAggregations.parse_obj(result.metrics) if result.metrics else None,
        )

    @deprecate_endpoint(
        path=f"{new_base_endpoint}/labeling/rules",
        new_path=f"{base_endpoint}/labeling/rules",
        router_method=router.get,
        operation_id="list_labeling_rules",
        description="List all dataset labeling rules",
        response_model=List[LabelingRule],
        response_model_exclude_none=True,
    )
    async def list_labeling_rules(
        name: str,
        common_params: CommonTaskHandlerDependencies = Depends(),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> List[LabelingRule]:
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TextClassificationDataset,
        )

        return [LabelingRule.parse_obj(rule) for rule in service.list_labeling_rules(dataset)]

    @deprecate_endpoint(
        path=f"{new_base_endpoint}/labeling/rules",
        new_path=f"{base_endpoint}/labeling/rules",
        router_method=router.post,
        operation_id="create_rule",
        description="Creates a new dataset labeling rule",
        response_model=LabelingRule,
        response_model_exclude_none=True,
    )
    async def create_rule(
        name: str,
        rule: CreateLabelingRule,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> LabelingRule:
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TextClassificationDataset,
        )

        rule = ServiceLabelingRule(
            **rule.dict(),
            author=current_user.username,
        )
        service.add_labeling_rule(
            dataset,
            rule=rule,
        )
        return LabelingRule.parse_obj(rule)

    @deprecate_endpoint(
        path=f"{new_base_endpoint}/labeling/rules/{{query:path}}/metrics",
        new_path=f"{base_endpoint}/labeling/rules/{{query:path}}/metrics",
        router_method=router.get,
        operation_id="compute_rule_metrics",
        description="Computes dataset labeling rule metrics",
        response_model=LabelingRuleMetricsSummary,
        response_model_exclude_none=True,
    )
    async def compute_rule_metrics(
        name: str,
        query: str,
        labels: Optional[List[str]] = Query(None, description="Label related to query rule", alias="label"),
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> LabelingRuleMetricsSummary:
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TextClassificationDataset,
        )

        return service.compute_labeling_rule(dataset, rule_query=query, labels=labels)

    @deprecate_endpoint(
        path=f"{new_base_endpoint}/labeling/rules/metrics",
        new_path=f"{base_endpoint}/labeling/rules/metrics",
        router_method=router.get,
        operation_id="compute_dataset_rules_metrics",
        description="Computes overall metrics for dataset labeling rules",
        response_model=DatasetLabelingRulesMetricsSummary,
        response_model_exclude_none=True,
    )
    async def compute_dataset_rules_metrics(
        name: str,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> DatasetLabelingRulesMetricsSummary:
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TextClassificationDataset,
        )
        metrics = service.compute_all_labeling_rules(dataset)
        return DatasetLabelingRulesMetricsSummary.parse_obj(metrics)

    @deprecate_endpoint(
        path=f"{new_base_endpoint}/labeling/rules/{{query:path}}",
        new_path=f"{base_endpoint}/labeling/rules/{{query:path}}",
        router_method=router.delete,
        operation_id="delete_labeling_rule",
        description="Deletes a labeling rule from dataset",
    )
    async def delete_labeling_rule(
        name: str,
        query: str,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> None:
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TextClassificationDataset,
        )

        service.delete_labeling_rule(dataset, rule_query=query)

    @deprecate_endpoint(
        path=f"{new_base_endpoint}/labeling/rules/{{query:path}}",
        new_path=f"{base_endpoint}/labeling/rules/{{query:path}}",
        router_method=router.get,
        operation_id="get_rule",
        description="Get the dataset labeling rule",
        response_model=LabelingRule,
        response_model_exclude_none=True,
    )
    async def get_rule(
        name: str,
        query: str,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> LabelingRule:
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TextClassificationDataset,
        )
        rule = service.find_labeling_rule(dataset, rule_query=query)
        return LabelingRule.parse_obj(rule)

    @deprecate_endpoint(
        path=f"{new_base_endpoint}/labeling/rules/{{query:path}}",
        new_path=f"{base_endpoint}/labeling/rules/{{query:path}}",
        router_method=router.patch,
        operation_id="update_rule",
        description="Update dataset labeling rule attributes",
        response_model=LabelingRule,
        response_model_exclude_none=True,
    )
    async def update_rule(
        name: str,
        query: str,
        update: UpdateLabelingRule,
        common_params: CommonTaskHandlerDependencies = Depends(),
        service: TextClassificationService = Depends(TextClassificationService.get_instance),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        current_user: User = Security(auth.get_current_user),
    ) -> LabelingRule:
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=task_type,
            workspace=common_params.workspace,
            as_dataset_class=TextClassificationDataset,
        )

        rule = service.update_labeling_rule(
            dataset,
            rule_query=query,
            labels=update.labels,
            description=update.description,
        )
        return LabelingRule.parse_obj(rule)

    text_classification_dataset_settings.configure_router(router)
    metrics.configure_router(
        router,
        cfg=TasksFactory.get_task_by_task_type(task_type),
    )
    return router


router = configure_router()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typing

from fastapi import Request
from fastapi.security import HTTPBearer
from starlette.authentication import AuthCredentials, AuthenticationBackend, BaseUser

from argilla_server.contexts import accounts
from argilla_server.security.authentication.jwt import JWT
from argilla_server.security.authentication.userinfo import UserInfo


class BearerTokenAuthenticationBackend(AuthenticationBackend):
    """Authenticate the user using the username and password Bearer header"""

    scheme = HTTPBearer(auto_error=False)

    async def authenticate(self, request: Request) -> typing.Optional[typing.Tuple[AuthCredentials, BaseUser]]:
        """Authenticate the user using the username and password Bearer header"""
        credentials = await self.scheme(request)
        if not credentials:
            return None

        token = credentials.credentials
        username = JWT.decode(token).get("username")

        db = request.state.db
        user = await accounts.get_user_by_username(db, username)
        if not user:
            return None

        return AuthCredentials(), UserInfo(
            username=user.username, name=user.first_name, role=user.role, identity=str(user.id)
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid

import pytest
from argilla_server.commons.models import TaskType
from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.daos.backend.search.model import BaseRecordsQuery
from argilla_server.errors import InvalidTextSearchError


@pytest.fixture(scope="session")
def engine():
    return GenericElasticEngineBackend.get_instance()


@pytest.fixture
def dataset_id(engine: GenericElasticEngineBackend):
    dataset_id = str(uuid.uuid4())

    yield dataset_id

    engine.delete(dataset_id)


def test_creating_index_with_non_searchable_metadata(engine: GenericElasticEngineBackend, dataset_id: str):
    engine.create_dataset(
        id=dataset_id,
        task=TaskType.text_classification,
        metadata_values={
            "a": "value",
            "other": "value",
            "_this": "is non searchable",
            "_other": "other disabled field",
        },
        force_recreate=True,
    )

    # Check the schema definition
    schema = engine.get_schema(dataset_id)
    assert schema["mappings"]["properties"]["metadata"]["properties"]["_this"] == {"type": "object", "enabled": False}
    assert schema["mappings"]["properties"]["metadata"]["properties"]["_other"] == {"type": "object", "enabled": False}


def test_non_searchable_docs_are_not_present_in_metrics(engine: GenericElasticEngineBackend, dataset_id: str):
    engine.create_dataset(
        id=dataset_id,
        task=TaskType.text_classification,
        metadata_values={"a": "value", "other": "value", "_this": "is non searchable"},
        force_recreate=True,
    )

    docs = [{"text": "This is my text", "metadata": {"_this": "value"}}] * 100
    assert engine.add_dataset_records(dataset_id, documents=docs) == 0

    metric_results = engine.compute_metric(dataset_id, metric_id="metadata")
    assert metric_results == {}


def test_non_searchable_fields_are_present_in_documents(engine: GenericElasticEngineBackend, dataset_id: str):
    engine.create_dataset(
        id=dataset_id,
        task=TaskType.text_classification,
        metadata_values={
            "a": "value",
            "other": "value",
            "_this": "is non searchable",
            "_other": "other disabled field",
        },
        force_recreate=True,
    )
    documents = [
        {"id": f"{i:03d}", "text": "This is my text", "metadata": {"_this": "value", "_other": {"with": "key"}}}
        for i in range(0, 100)
    ]

    assert engine.add_dataset_records(dataset_id, documents=documents) == 0

    total, results = engine.search_records(dataset_id, size=5)
    assert total == 100
    assert results == documents[:5]


def test_non_searchable_fields_cannot_be_used_for_search(engine: GenericElasticEngineBackend, dataset_id: str):
    engine.create_dataset(
        id=dataset_id,
        task=TaskType.text_classification,
        metadata_values={"a": "value", "_protected": "is non searchable", "non_protected": "normal field"},
        force_recreate=True,
    )
    documents = [
        {"id": f"{i:03d}", "text": "This is my text", "metadata": {"_protected": "value", "non_protected": "value"}}
        for i in range(0, 100)
    ]

    assert engine.add_dataset_records(dataset_id, documents=documents) == 0

    total, _ = engine.search_records(id=dataset_id, query=BaseRecordsQuery(query_text="metadata.non_protected:value"))
    assert total == 100

    total, _ = engine.search_records(id=dataset_id, query=BaseRecordsQuery(query_text="metadata._protected:value"))
    assert total == 0



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.errors.api_errors import APIErrorHandler
from argilla_server.errors.base_errors import (
    EntityAlreadyExistsError,
    EntityNotFoundError,
    GenericServerError,
    ServerError,
)
from argilla_server.schemas.v0.datasets import Dataset
from fastapi import Request

mock_request = Request(scope={"type": "http", "headers": {}})


@pytest.mark.asyncio
class TestAPIErrorHandler:
    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["error", "expected_event"],
        [
            (
                EntityNotFoundError(name="mock-name", type="MockType"),
                {
                    "accept-language": None,
                    "code": "argilla.api.errors::EntityNotFoundError",
                    "type": "MockType",
                    "user-agent": None,
                },
            ),
            (
                EntityAlreadyExistsError(name="mock-name", type=Dataset, workspace="mock-workspace"),
                {
                    "accept-language": None,
                    "code": "argilla.api.errors::EntityAlreadyExistsError",
                    "type": "Dataset",
                    "user-agent": None,
                },
            ),
            (
                GenericServerError(RuntimeError("This is a mock error")),
                {
                    "accept-language": None,
                    "code": "argilla.api.errors::GenericServerError",
                    "type": "builtins.RuntimeError",
                    "user-agent": None,
                },
            ),
            (
                ServerError(),
                {
                    "accept-language": None,
                    "code": "argilla.api.errors::ServerError",
                    "user-agent": None,
                },
            ),
        ],
    )
    async def test_track_error(self, test_telemetry, error, expected_event):
        await APIErrorHandler.track_error(error, request=mock_request)

        test_telemetry.track_data.assert_called_once_with(action="ServerErrorFound", data=expected_event)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer


def start(
    host: str = typer.Option("0.0.0.0", help="The host where the Argilla server will be binded"),
    port: int = typer.Option(6900, help="The port where the Argilla server will be binded"),
    access_log: bool = typer.Option(True, help="Whether to enable or disable the Argilla server access log"),
) -> None:
    import uvicorn

    uvicorn.run(
        "argilla_server:app",
        port=port,
        host=host,
        access_log=access_log,
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import asyncio
from typing import AsyncGenerator, Optional
from uuid import UUID

import typer
from rich.progress import Progress
from sqlalchemy import func, select
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload

from argilla_server.database import AsyncSessionLocal
from argilla_server.models import Dataset, Record, Response, Suggestion
from argilla_server.search_engine import SearchEngine, get_search_engine

from .rich import echo_in_panel


class Reindexer:
    YIELD_PER = 100

    @classmethod
    async def reindex_dataset(cls, db: AsyncSession, search_engine: SearchEngine, dataset_id: UUID) -> Dataset:
        dataset = (
            await db.execute(
                select(Dataset)
                .filter_by(id=dataset_id)
                .options(
                    selectinload(Dataset.fields),
                    selectinload(Dataset.questions),
                    selectinload(Dataset.metadata_properties),
                    selectinload(Dataset.vectors_settings),
                )
            )
        ).scalar_one()

        await search_engine.delete_index(dataset)
        await search_engine.create_index(dataset)

        return dataset

    @classmethod
    async def reindex_datasets(cls, db: AsyncSession, search_engine: SearchEngine) -> AsyncGenerator[Dataset, None]:
        stream = await db.stream(
            select(Dataset)
            .order_by(Dataset.inserted_at.asc())
            .options(
                selectinload(Dataset.fields),
                selectinload(Dataset.questions),
                selectinload(Dataset.metadata_properties),
                selectinload(Dataset.vectors_settings),
            )
            .execution_options(yield_per=cls.YIELD_PER)
        )

        async for dataset in stream.scalars():
            await search_engine.delete_index(dataset)
            await search_engine.create_index(dataset)

            yield dataset

    @classmethod
    async def reindex_dataset_records(
        cls, db: AsyncSession, search_engine: SearchEngine, dataset: Dataset
    ) -> AsyncGenerator[list[Record], None]:
        stream = await db.stream(
            select(Record)
            .filter_by(dataset_id=dataset.id)
            .order_by(Record.inserted_at.asc())
            .options(
                selectinload(Record.responses).selectinload(Response.user),
                selectinload(Record.suggestions).selectinload(Suggestion.question),
                selectinload(Record.vectors),
            )
            .execution_options(yield_per=cls.YIELD_PER)
        )

        async for records_partition in stream.partitions():
            records = [record for (record,) in records_partition]

            await search_engine.index_records(dataset, records)

            yield records

    @classmethod
    async def count_datasets(cls, db: AsyncSession) -> int:
        return (await db.execute(select(func.count(Dataset.id)))).scalar()

    @classmethod
    async def count_dataset_records(cls, db: AsyncSession, dataset: Dataset) -> int:
        return (await db.execute(select(func.count(Record.id)).filter_by(dataset_id=dataset.id))).scalar()


async def _reindex_dataset(db: AsyncSession, search_engine: SearchEngine, progress: Progress, dataset_id: UUID) -> None:
    try:
        dataset = await Reindexer.reindex_dataset(db, search_engine, dataset_id)
    except NoResultFound as e:
        echo_in_panel(
            f"Feedback dataset with id={dataset_id} not found.",
            title="Feedback dataset not found",
            title_align="left",
            success=False,
        )

        raise typer.Exit(code=1) from e

    task = progress.add_task(f"reindexing feedback dataset `{dataset.name}`...", total=1)

    await _reindex_dataset_records(db, search_engine, progress, dataset)

    progress.advance(task)


async def _reindex_datasets(db: AsyncSession, search_engine: SearchEngine, progress: Progress) -> None:
    task = progress.add_task(f"reindexing feedback datasets...", total=await Reindexer.count_datasets(db))

    async for dataset in Reindexer.reindex_datasets(db, search_engine):
        await _reindex_dataset_records(db, search_engine, progress, dataset)

        progress.advance(task)


async def _reindex_dataset_records(
    db: AsyncSession, search_engine: SearchEngine, progress: Progress, dataset: Dataset
) -> None:
    task = progress.add_task(
        f"reindexing feedback dataset `{dataset.name}` records...",
        total=await Reindexer.count_dataset_records(db, dataset),
    )

    async for records in Reindexer.reindex_dataset_records(db, search_engine, dataset):
        progress.advance(task, advance=len(records))


async def _reindex(feedback_dataset_id: Optional[UUID] = None) -> None:
    async with AsyncSessionLocal() as db:
        async for search_engine in get_search_engine():
            with Progress() as progress:
                if feedback_dataset_id is not None:
                    await _reindex_dataset(db, search_engine, progress, feedback_dataset_id)
                else:
                    await _reindex_datasets(db, search_engine, progress)


def reindex(
    feedback_dataset_id: Optional[UUID] = typer.Option(None, help="The id of a feedback dataset to be reindexed")
) -> None:
    asyncio.run(_reindex(feedback_dataset_id))


if __name__ == "__main__":
    typer.run(reindex)
